# #
# Software subject to following license(s):
#   The Apache Software License, Version 2.0 (http://www.apache.org/licenses/LICENSE-2.0.txt)
#   null
#

# ${developer-info
# ${author-info}
# #
# pxelinux, 14.5.0-rc10, 20140606.1133.15
#

package NCM::Component::pxelinux;

use strict;
use warnings;
use NCM::Component;
use EDG::WP4::CCM::Property;
use NCM::Check;
use Sys::Hostname;
use CAF::FileWriter;
use NCM::Component::ks qw (ksuserhooks);
use LC::Fatal qw (symlink);
use File::stat;
use Time::localtime;

use constant PXEROOT => "/system/aii/nbp/pxelinux";
use constant NBPDIR => 'nbpdir';
use constant LOCALBOOT => 'bootconfig';
use constant HOSTNAME => "/system/network/hostname";
use constant DOMAINNAME => "/system/network/domainname";
use constant ETH => "/system/network/interfaces";
use constant INSTALL => 'install';
use constant BOOT => 'boot';
use constant RESCUE => 'rescue';
use constant RESCUEBOOT => 'rescueconfig';
use constant FIRMWARE => 'firmware';
use constant LIVECD => 'livecd';
# Hooks for NBP plug-in
use constant RESCUE_HOOK_PATH => '/system/aii/hooks/rescue';
use constant INSTALL_HOOK_PATH => '/system/aii/hooks/install';
use constant REMOVE_HOOK_PATH => '/system/aii/hooks/remove';
use constant BOOT_HOOK_PATH => '/system/aii/hooks/boot';
use constant FIRMWARE_HOOK_PATH => '/system/aii/hooks/firmware';
use constant LIVECD_HOOK_PATH => '/system/aii/hooks/livecd';
# Kickstart constants (trying to use same name as in ks.pm from aii-ks)
use constant KS => "/system/aii/osinstall/ks";

our @ISA = qw (NCM::Component);
our $EC = LC::Exception::Context->new->will_store_all;
our $this_app = $main::this_app;

# Returns the absolute path where the PXE file must be written.
sub filepath
{
    my $cfg = shift;

    my $h = $cfg->getElement (HOSTNAME)->getValue;
    my $d = $cfg->getElement (DOMAINNAME)->getValue;
    my $dir = $this_app->option (NBPDIR);
    $this_app->debug(3, "NBP directory = $dir");
    return "$dir/$h.$d.cfg";
}

# Returns the absolute path of the PXE file to link to
sub link_filepath
{
    my ($cfg, $cmd) = @_;

    my $dir = $this_app->option (NBPDIR);

    my $cfgpath = PXEROOT . "/" . $cmd;
    if ($cfg->elementExists ($cfgpath)) {
        my $linkname = $cfg->getElement ($cfgpath)->getValue;
        return "$dir/$linkname";
    } elsif ($cmd eq RESCUE) {
        # Backwards compatibility: use the option set on the command line
        # if the profile does not define a rescue image
        my $path = $this_app->option (RESCUEBOOT);
        unless ($path =~ m{^([-.\w]+)$}) {
            $this_app->error ("Unexpected RESCUE configuration file");
        }
        return "$dir/$1";
    } else {
        my $h = $cfg->getElement (HOSTNAME)->getValue;
        my $d = $cfg->getElement (DOMAINNAME)->getValue;
        $this_app->debug(3, "No $cmd defined for $h.$d");
    }
    return undef;
}

# create a list with all append options
sub pxe_append 
{
    my $cfg = shift;

    my $t = $cfg->getElement (PXEROOT)->getTree;
    
    my $kst = {}; # empty hashref in case no kickstart is defined
    $kst = $cfg->getElement (KS)->getTree if $cfg->elementExists(KS);

    my $ksloc = $t->{kslocation};
    my $server = hostname();
    $ksloc =~ s{LOCALHOST}{$server};

    my @append;
    push(@append,
         "ramdisk=32768",
         "initrd=$t->{initrd}",
         "ks=$ksloc",
         "ksdevice=$t->{ksdevice}"
         );         

    if (exists($kst->{logging})) {
        push(@append, "syslog=$kst->{logging}->{host}:$kst->{logging}->{port}"); 
        push(@append, "loglevel=$kst->{logging}->{level}") if $kst->{logging}->{level};
    }
        
    push(@append, $t->{append}) if exists $t->{append};

    return @append;    
}

# Prints the PXE configuration file.
sub pxeprint
{
    my $cfg = shift;
    my $t = $cfg->getElement (PXEROOT)->getTree;
    my $fh = CAF::FileWriter->open (filepath ($cfg),
				    log => $this_app, mode => 0644);

    my $appendtxt = join(" ", pxe_append($cfg));

    $fh->print (<<EOF
# File generated by pxelinux AII plug-in.
# Do not edit.
default $t->{label}
    label $t->{label}
    kernel $t->{kernel}
    append $appendtxt
EOF
           );

    $fh->print ("    ipappend 2\n") if $t->{ksdevice} eq 'bootif';
    $fh->close();
}

# Prints an IP address in hexadecimal.
sub hexip
{
    my $ip = shift || "";

    return sprintf ("%02X%02X%02X%02X", $1, $2, $3, $4) if ($ip =~ m/^(\d+)\.(\d+)\.(\d+)\.(\d+)$/);
}

# Creates a symbolic link for PXE. This means creating a symlink named
# after the node's IP in hexadecimal to a PXE file.
sub pxelink
{
    my ($cfg, $cmd) = @_;

    my $t = $cfg->getElement (ETH)->getTree;
    my $path;
    if (!$cmd) {
        $path = $this_app->option (LOCALBOOT);
        $this_app->debug (5, "Configuring on $path");
    } elsif ($cmd eq BOOT) {
        $path = $this_app->option (LOCALBOOT);
        unless ($path =~ m{^([-.\w]+)$}) {
            $this_app->error ("Unexpected BOOT configuration file");
            return -1;
        }
        $path = $1;
        $this_app->debug (5, "Local booting from $path");
    } elsif ($cmd eq RESCUE || $cmd eq LIVECD || $cmd eq FIRMWARE) {
        $path = link_filepath($cfg, $cmd);
        if (! -f $path) {
            my $h = $cfg->getElement (HOSTNAME)->getValue;
            my $d = $cfg->getElement (DOMAINNAME)->getValue;
            $this_app->error("Missing $cmd config file for $h.$d: $path");
            return -1;
        }
        $this_app->debug (5, "Using $cmd from: $path");
    } elsif ($cmd eq INSTALL) {
        $path = filepath ($cfg);
        $this_app->debug (5, "Installing on $path");
    } else {
        $this_app->debug (5, "Unknown command");
        return -1;
    }
    # Set the same settings for every network interface that has a
    # defined IP address.
    foreach my $st (values (%$t)) {
        next unless exists ($st->{ip});
        my $dir = $this_app->option (NBPDIR);
        my $lnname = "$dir/".hexip ($st->{ip});
        if ($cmd || ! -l $lnname) {
            if ($NoAction) {
                $this_app->info ("Would symlink $path to $lnname");
            } else {
                unlink ($lnname);
                # This must be stripped to work with chroot'edg
                # environments.
                $path =~ s{$dir/?}{};
                symlink ($path, $lnname);
            }
        }
    }
    return 0;
}

# Sets the node's status to install.
sub Install
{
    my ($self, $cfg) = @_;

    unless (pxelink ($cfg, INSTALL)==0) {
        my $h = $cfg->getElement (HOSTNAME)->getValue;
        my $d = $cfg->getElement (DOMAINNAME)->getValue;
        $self->error ("Failed to change the status of $h.$d to install");
        return 0;
    }
    ksuserhooks ($cfg, INSTALL_HOOK_PATH) unless $NoAction;
    return 1;
}

# Sets the node's status to firmware
sub Firmware
{
    my ($self, $cfg) = @_;

    unless (pxelink ($cfg, FIRMWARE)==0) {
        my $h = $cfg->getElement (HOSTNAME)->getValue;
        my $d = $cfg->getElement (DOMAINNAME)->getValue;
        $self->error ("Failed to change the status of $h.$d to firmware");
        return 0;
    }
    ksuserhooks ($cfg, FIRMWARE_HOOK_PATH) unless $NoAction;
    return 1;
}

# Sets the node's status to livecd
sub Livecd
{
    my ($self, $cfg) = @_;

    unless (pxelink ($cfg, LIVECD)==0) {
        my $h = $cfg->getElement (HOSTNAME)->getValue;
        my $d = $cfg->getElement (DOMAINNAME)->getValue;
        $self->error("Failed to change the status of $h.$d to livecd");
        return 0;
    }
    ksuserhooks ($cfg, LIVECD_HOOK_PATH) unless $NoAction;
    return 1;
}

# Sets the node's status to rescue.
sub Rescue
{
    my ($self, $cfg) = @_;

    unless (pxelink ($cfg, RESCUE)==0) {
        my $h = $cfg->getElement (HOSTNAME)->getValue;
        my $d = $cfg->getElement (DOMAINNAME)->getValue;
        $self->error ("Failed to change the status of $h.$d to rescue");
        return 0;
    }
    ksuserhooks ($cfg, RESCUE_HOOK_PATH) unless $NoAction;
    return 1;
}

# Prints the status of the node.
sub Status
{
    my ($self, $cfg) = @_;

    my $t = $cfg->getElement (ETH)->getTree;
    my $dir = $this_app->option (NBPDIR);
    my $h = $cfg->getElement (HOSTNAME)->getValue;
    my $d = $cfg->getElement (DOMAINNAME)->getValue;
    my $fqdn = "$h.$d";
    my $boot = $this_app->option (LOCALBOOT);
    my $rescue = link_filepath($cfg, RESCUE);
    my $firmware = link_filepath($cfg, FIRMWARE);
    my $livecd = link_filepath($cfg, LIVECD);
    foreach my $s (values (%$t)) {
        next unless exists ($s->{ip});
        my $ln = hexip ($s->{ip});
        my $since = "unknown";
        my $st;
        if (-l "$dir/$ln") {
            $since = ctime(lstat("$dir/$ln")->ctime());
            my $name = readlink ("$dir/$ln");
            my $name_path = "$dir/$name";
            if (! -e $name_path) {
                $st = "broken";
            } elsif ($name =~ m{^(?:.*/)?$fqdn\.cfg$}) {
                $st = "install";
            } elsif ($name =~ m{^$boot$}) {
                $st = "boot";
            } elsif ($firmware && ($name_path =~ m{$firmware})) {
                $st = "firmware";
            } elsif ($livecd && ($name_path =~ m{$livecd})) {
                $st = "livecd";
            } elsif ($rescue && ($name_path =~ m{$rescue})) {
                $st = "rescue";
            } else {
                $st = "unknown";
            }
        } else {
            $st = "undefined";
        }
        $self->info(ref($self), " status for $fqdn: $s->{ip} $st ",
                "since: $since");
    }
    return 1;
}

# Sets the node's status to boot from local boot.
sub Boot
{
    my ($self, $cfg) = @_;
    pxelink ($cfg, BOOT);
    ksuserhooks ($cfg, BOOT_HOOK_PATH) unless $NoAction;
    return 1;
}

# Creates the PXE configuration file.
sub Configure
{
    my ($self, $cfg) = @_;

    pxeprint ($cfg);
    pxelink ($cfg);

    return 1;
}

# Removes PXE files and symlinks for the node. To be called by --remove.
sub Unconfigure
{
    my ($self, $cfg) = @_;

    if ($NoAction) {
        $self->info ("Would remove " . ref ($self));
        return 1;
    }

    my $t = $cfg->getElement (ETH)->getTree;
    my $path = filepath ($cfg);
    my $dir = $this_app->option (NBPDIR);
    # Set the same settings for every network interface.
    unlink ($path);
    unlink ("$dir/" . hexip ($_->{ip})) foreach values (%$t);
    ksuserhooks ($cfg, REMOVE_HOOK_PATH);
    return 1;
}
